@@ -32,7 +32,6 @@ from oslo_utils import encodeutils
 from oslo_utils import importutils
 from oslo_utils import strutils
 from oslo_utils import timeutils
-import six
 
 from oslo_concurrency._i18n import _
 
@@ -56,6 +55,13 @@ if eventlet_patched:
         from eventlet.green import subprocess
 
     from eventlet import tpool
+
+    # Monkey patch the original current_thread to use the up-to-date _active
+    # global variable. See https://bugs.launchpad.net/bugs/1863021 and
+    # https://github.com/eventlet/eventlet/issues/592
+    import __original_module_threading as orig_threading
+    import threading
+    orig_threading.current_thread.__globals__['_active'] = threading._active
 else:
     import subprocess
 
@@ -174,7 +180,7 @@ class ProcessLimits(object):
     }
 
     def __init__(self, **kw):
-        for limit in self._LIMITS.keys():
+        for limit in self._LIMITS:
             setattr(self, limit, kw.pop(limit, None))
 
         if kw:
@@ -184,7 +190,7 @@ class ProcessLimits(object):
     def prlimit_args(self):
         """Create a list of arguments for the prlimit command line."""
         args = []
-        for limit in self._LIMITS.keys():
+        for limit in self._LIMITS:
             val = getattr(self, limit)
             if val is not None:
                 args.append("%s=%s" % (self._LIMITS[limit], val))
@@ -263,14 +269,19 @@ def execute(*cmd, **kwargs):
                             below for a detailed description.
     :type prlimit:          :class:`ProcessLimits`
     :param python_exec:     The python executable to use for enforcing
-                            prlimits. If this is not set it will default to use
-                            sys.executable.
+                            prlimits. If this is not set or is None, it will
+                            default to use sys.executable.
     :type python_exec:      string
+    :param timeout:         Timeout (in seconds) to wait for the process
+                            termination. If timeout is reached,
+                            :class:`subprocess.TimeoutExpired` is raised.
+    :type timeout:          int
     :returns:               (stdout, stderr) from process execution
     :raises:                :class:`UnknownArgumentError` on
                             receiving unknown arguments
     :raises:                :class:`ProcessExecutionError`
     :raises:                :class:`OSError`
+    :raises:                :class:`subprocess.TimeoutExpired`
 
     The *prlimit* parameter can be used to set resource limits on the child
     process.  If this parameter is used, the child process will be spawned by a
@@ -318,7 +329,8 @@ def execute(*cmd, **kwargs):
     on_completion = kwargs.pop('on_completion', None)
     preexec_fn = kwargs.pop('preexec_fn', None)
     prlimit = kwargs.pop('prlimit', None)
-    python_exec = kwargs.pop('python_exec', sys.executable)
+    python_exec = kwargs.pop('python_exec', None) or sys.executable
+    timeout = kwargs.pop('timeout', None)
 
     if isinstance(check_exit_code, bool):
         ignore_exit_code = not check_exit_code
@@ -329,7 +341,7 @@ def execute(*cmd, **kwargs):
     if kwargs:
         raise UnknownArgumentError(_('Got unknown keyword args: %r') % kwargs)
 
-    if isinstance(log_errors, six.integer_types):
+    if isinstance(log_errors, int):
         log_errors = LogErrors(log_errors)
     if not isinstance(log_errors, LogErrors):
         raise InvalidArgumentError(_('Got invalid arg log_errors: %r') %
@@ -386,7 +398,7 @@ def execute(*cmd, **kwargs):
                                    stderr=_PIPE,
                                    close_fds=close_fds,
                                    preexec_fn=on_preexec_fn,
-                                   shell=shell,
+                                   shell=shell,  # nosec:B604
                                    cwd=cwd,
                                    env=env_variables)
 
@@ -399,30 +411,35 @@ def execute(*cmd, **kwargs):
                 # we have to wrap this call using tpool.
                 if eventlet_patched and os.name == 'nt':
                     result = tpool.execute(obj.communicate,
-                                           process_input)
+                                           process_input,
+                                           timeout=timeout)
                 else:
-                    result = obj.communicate(process_input)
+                    result = obj.communicate(process_input,
+                                             timeout=timeout)
 
                 obj.stdin.close()  # pylint: disable=E1101
                 _returncode = obj.returncode  # pylint: disable=E1101
                 LOG.log(loglevel, 'CMD "%s" returned: %s in %0.3fs',
                         sanitized_cmd, _returncode, watch.elapsed())
+            except subprocess.TimeoutExpired:
+                LOG.log(loglevel, 'CMD "%s" reached timeout in %0.3fs',
+                        sanitized_cmd, watch.elapsed())
+                raise
             finally:
                 if on_completion:
                     on_completion(obj)
 
             if not ignore_exit_code and _returncode not in check_exit_code:
                 (stdout, stderr) = result
-                if six.PY3:
-                    stdout = os.fsdecode(stdout)
-                    stderr = os.fsdecode(stderr)
+                stdout = os.fsdecode(stdout)
+                stderr = os.fsdecode(stderr)
                 sanitized_stdout = strutils.mask_password(stdout)
                 sanitized_stderr = strutils.mask_password(stderr)
                 raise ProcessExecutionError(exit_code=_returncode,
                                             stdout=sanitized_stdout,
                                             stderr=sanitized_stderr,
                                             cmd=sanitized_cmd)
-            if six.PY3 and not binary and result is not None:
+            if not binary and result is not None:
                 (stdout, stderr) = result
                 # Decode from the locale using using the surrogateescape error
                 # handler (decoding cannot fail)
@@ -491,7 +508,7 @@ def trycmd(*args, **kwargs):
         out, err = execute(*args, **kwargs)
         failed = False
     except ProcessExecutionError as exn:
-        out, err = '', six.text_type(exn)
+        out, err = '', str(exn)
         failed = True
 
     if not failed and discard_warnings and err:
@@ -503,9 +520,21 @@ def trycmd(*args, **kwargs):
 
 def ssh_execute(ssh, cmd, process_input=None,
                 addl_env=None, check_exit_code=True,
-                binary=False, timeout=None):
+                binary=False, timeout=None,
+                sanitize_stdout=True):
     """Run a command through SSH.
 
+    :param ssh:             An SSH Connection object.
+    :param cmd:             The command string to run.
+    :param check_exit_code: If an exception should be raised for non-zero
+                            exit.
+    :param timeout:         Max time in secs to wait for command execution.
+    :param sanitize_stdout: Defaults to True. If set to True, stdout is
+                            sanitized i.e. any sensitive information like
+                            password in command output will be masked.
+    :returns:               (stdout, stderr) from command execution through
+                            SSH.
+
     .. versionchanged:: 1.9
        Added *binary* optional parameter.
     """
@@ -531,38 +560,35 @@ def ssh_execute(ssh, cmd, process_input=None,
 
     exit_status = channel.recv_exit_status()
 
-    if six.PY3:
-        # Decode from the locale using using the surrogateescape error handler
-        # (decoding cannot fail). Decode even if binary is True because
-        # mask_password() requires Unicode on Python 3
-        stdout = os.fsdecode(stdout)
-        stderr = os.fsdecode(stderr)
-    stdout = strutils.mask_password(stdout)
+    # Decode from the locale using using the surrogateescape error handler
+    # (decoding cannot fail). Decode even if binary is True because
+    # mask_password() requires Unicode on Python 3
+    stdout = os.fsdecode(stdout)
+    stderr = os.fsdecode(stderr)
+
+    if sanitize_stdout:
+        stdout = strutils.mask_password(stdout)
+
     stderr = strutils.mask_password(stderr)
 
     # exit_status == -1 if no exit code was returned
     if exit_status != -1:
         LOG.debug('Result was %s' % exit_status)
         if check_exit_code and exit_status != 0:
+            # In case of errors in command run, due to poor implementation of
+            # command executable program, there might be chance that it leaks
+            # sensitive information like password to stdout. In such cases
+            # stdout needs to be sanitized even though sanitize_stdout=False.
+            stdout = strutils.mask_password(stdout)
             raise ProcessExecutionError(exit_code=exit_status,
                                         stdout=stdout,
                                         stderr=stderr,
                                         cmd=sanitized_cmd)
 
     if binary:
-        if six.PY2:
-            # On Python 2, stdout is a bytes string if mask_password() failed
-            # to decode it, or an Unicode string otherwise. Encode to the
-            # default encoding (ASCII) because mask_password() decodes from
-            # the same encoding.
-            if isinstance(stdout, unicode):
-                stdout = stdout.encode()
-            if isinstance(stderr, unicode):
-                stderr = stderr.encode()
-        else:
-            # fsencode() is the reverse operation of fsdecode()
-            stdout = os.fsencode(stdout)
-            stderr = os.fsencode(stderr)
+        # fsencode() is the reverse operation of fsdecode()
+        stdout = os.fsencode(stdout)
+        stderr = os.fsencode(stderr)
 
     return (stdout, stderr)
 
